// SPDX-FileCopyrightText: 2012 Icinga GmbH <https://icinga.com>
// SPDX-License-Identifier: GPL-2.0-or-later

#ifndef I2LIVESTATUS_H
#define I2LIVESTATUS_H

/**
 * @defgroup icinga Livestatus
 *
 * The Livestatus library implements the Livestatus protocol for Icinga.
 */

#include "base/i2-base.hpp"

#endif /* I2LIVESTATUS_H */
