// SPDX-FileCopyrightText: 2012 Icinga GmbH <https://icinga.com>
// SPDX-License-Identifier: GPL-2.0-or-later

#include "base/boolean.hpp"
#include "base/primitivetype.hpp"

using namespace icinga;

REGISTER_BUILTIN_TYPE(Boolean, Boolean::GetPrototype());
